// Arduino G-code Interpreter
// v1.0 by Mike Ellery - initial software (mellery@gmail.com)
// v1.1 by Zach Smith - cleaned up and did lots of tweaks (hoeken@gmail.com)
// v1.2 by Chris Meighan - cleanup / G2&G3 support (cmeighan@gmail.com)
// v1.3 by Zach Smith - added thermocouple support and multi-sample temp readings. (hoeken@gmail.com)
//      also added preliminary and experimental support for rotary encoders.
// v1.3.1 by Wade Bortz - minor tweaks to rotary encoder code, dwell, etc (wbortz@gmail.com)
// v1.3.2 by Wade Bortz - hacked extruder_manage_temperature() to check for broken thermistor (wbortz@gmail.com)
#include <HardwareSerial.h>

//our command string
#define COMMAND_SIZE 128
#include "WProgram.h"
void setup();
void loop();
void delayMicrosecondsInterruptible(unsigned int us);
void extruder_read_quadrature();
void init_extruder();
void extruder_set_temperature(int temp);
int extruder_get_temperature();
int extruder_read_thermistor();
int extruder_read_thermocouple();
int extruder_sample_temperature(byte pin);
void extruder_manage_temperature();
void init_process_string();
void process_string(char instruction[], int size);
double search_string(char key, char instruction[], int string_size);
bool has_command(char key, char instruction[], int string_size);
void figure_feedrate(char instruction[], int size, byte code);
void init_steppers();
void dda_move(long micro_delay);
bool can_step(byte min_pin, byte max_pin, long current, long target, byte direction);
void do_step(byte step_pin);
bool read_switch(byte pin);
long to_steps(float steps_per_unit, float units);
void set_target(float x, float y, float z);
void set_position(float x, float y, float z);
void calculate_deltas();
long calculate_feedrate_delay(float feedrateW);
long getMaxSpeed();
void disable_steppers();
void enableTimer1Interrupt();
void disableTimer1Interrupt();
void setTimer1Resolution(byte r);
void setTimer1Ceiling(unsigned int c);
unsigned int getTimer1Ceiling(unsigned long ticks);
byte getTimer1Resolution(unsigned long ticks);
void setTimer1Ticks(unsigned long ticks);
void setupTimer1Interrupt();
void enableTimer2Interrupt();
void disableTimer2Interrupt();
void setTimer2Resolution(byte r);
void setTimer2Ceiling(byte c);
byte getTimer2Ceiling(unsigned long ticks);
byte getTimer2Resolution(unsigned long ticks);
void setTimer2Ticks(unsigned long ticks);
void setupTimer2Interrupt();
char word[COMMAND_SIZE];
byte serial_count;
int no_data = 0;

void setup()
{
	//Do startup stuff here
	Serial.begin(19200);
	Serial.println("start");
	Serial.println("v1.3.1");
	
	//other initialization.
	init_process_string();
	init_steppers();
	init_extruder();
}

void loop()
{
	char c;
	
	//keep it hot!
	extruder_manage_temperature();

	//read in characters if we got them.
	if (Serial.available() > 0)
	{
		c = Serial.read();
		no_data = 0;
		
		//newlines are ends of commands.
		if (c != '\n')
		{
			word[serial_count] = c;
			serial_count++;
		}
	}
	//mark no data.
	else
	{
		no_data++;
		delayMicrosecondsInterruptible(100);
	}

	//if theres a pause or we got a real command, do it
	if (serial_count && (c == '\n' || no_data > 100))
	{
		//process our command!
		process_string(word, serial_count);

		//clear command.
		init_process_string();
	}

	//no data?  turn off steppers
	if (no_data > 1000)
		disable_steppers();
}

void delayMicrosecondsInterruptible(unsigned int us)
{

#if F_CPU >= 16000000L
    // for the 16 MHz clock on most Arduino boards

	// for a one-microsecond delay, simply return.  the overhead
	// of the function call yields a delay of approximately 1 1/8 us.
	if (--us == 0)
		return;

	// the following loop takes a quarter of a microsecond (4 cycles)
	// per iteration, so execute it four times for each microsecond of
	// delay requested.
	us <<= 2;

	// account for the time taken in the preceeding commands.
	us -= 2;
#else
    // for the 8 MHz internal clock on the ATmega168

    // for a one- or two-microsecond delay, simply return.  the overhead of
    // the function calls takes more than two microseconds.  can't just
    // subtract two, since us is unsigned; we'd overflow.
	if (--us == 0)
		return;
	if (--us == 0)
		return;

	// the following loop takes half of a microsecond (4 cycles)
	// per iteration, so execute it twice for each microsecond of
	// delay requested.
	us <<= 1;
    
    // partially compensate for the time taken by the preceeding commands.
    // we can't subtract any more than this or we'd overflow w/ small delays.
    us--;
#endif

	// busy wait
	__asm__ __volatile__ (
		"1: sbiw %0,1" "\n\t" // 2 cycles
		"brne 1b" : "=w" (us) : "0" (us) // 2 cycles
	);
}

// define the parameters of our machine.
#define X_STEPS_PER_INCH 203.2
#define X_STEPS_PER_MM   8
#define X_MOTOR_STEPS    400

#define Y_STEPS_PER_INCH 203.2
#define Y_STEPS_PER_MM   8
#define Y_MOTOR_STEPS    400

#define Z_STEPS_PER_INCH 8128
#define Z_STEPS_PER_MM   320
#define Z_MOTOR_STEPS    400

//our maximum feedrates (mm/min?)
#define FAST_XY_FEEDRATE 1000.0
#define FAST_Z_FEEDRATE  200.0

// Units in curve section
#define CURVE_SECTION_INCHES 0.019685
#define CURVE_SECTION_MM 0.5

// Set to one if sensor outputs inverting (ie: 1 means open, 0 means closed)
// RepRap opto endstops are *not* inverting.
#define SENSORS_INVERTING 0

// How many temperature samples to take.  each sample takes about 100 usecs.
#define TEMPERATURE_SAMPLES 5

//these defines are for using rotary encoders on the extruder
#define EXTRUDER_ENCODER_STEPS 512		//number of steps per revolution
#define EXTRUDER_MIN_SPEED 50			//minimum PWM speed to use
#define EXTRUDER_MAX_SPEED 255			//maximum PWM speed to use
#define EXTRUDER_ERROR_MARGIN 10		//our error margin (to prevent constant seeking)

//here are our defines for default extruder PID values
#define EXTRUDER_INITIAL_PGAIN 5
#define EXTRUDER_INITIAL_IGAIN 10
#define EXTRUDER_INITIAL_DGAIN 10

/****************************************************************************************
* digital i/o pin assignment
*
* this uses the undocumented feature of Arduino - pins 14-19 correspond to analog 0-5
****************************************************************************************/
//cartesian bot pins  Updated for my setup - WRB
#define X_STEP_PIN 9
#define X_DIR_PIN 12
#define X_MIN_PIN 15
#define X_MAX_PIN 20
#define X_ENABLE_PIN 20

#define Y_STEP_PIN 10
#define Y_DIR_PIN 7
#define Y_MIN_PIN 8
#define Y_MAX_PIN 20
#define Y_ENABLE_PIN 20

#define Z_STEP_PIN 19
#define Z_DIR_PIN 18
#define Z_MIN_PIN 17
#define Z_MAX_PIN 20
#define Z_ENABLE_PIN 20

//extruder pins
#define EXTRUDER_ENCODER_A_PIN 2 		//quadrature a pin
#define EXTRUDER_ENCODER_B_PIN 3		//quadrature b pin
#define EXTRUDER_MOTOR_DIR_PIN     4
#define EXTRUDER_MOTOR_SPEED_PIN  5
#define EXTRUDER_HEATER_PIN        6
#define EXTRUDER_FAN_PIN           11
#define EXTRUDER_THERMISTOR_PIN    -1  //a -1 disables thermistor readings
#define EXTRUDER_THERMOCOUPLE_PIN  0 //a -1 disables thermocouple readings

#define INVERT_EXTRUDER_QUADRATURE  //the MRE is inverted.

#define EXTRUDER_VALVE_PIN         -1 //this wasn't defined, which tells me that I've got missmatched files from Sourceforge.  Uh oh.  WRB
/*
//cartesian bot pins
#define X_STEP_PIN 7
#define X_DIR_PIN 8
#define X_MIN_PIN 14
#define X_MAX_PIN 17
#define X_ENABLE_PIN 18

#define Y_STEP_PIN 9
#define Y_DIR_PIN 10
#define Y_MIN_PIN 15
#define Y_MAX_PIN 17
#define Y_ENABLE_PIN 18

#define Z_STEP_PIN 12
#define Z_DIR_PIN 13
#define Z_MIN_PIN 16
#define Z_MAX_PIN 17
#define Z_ENABLE_PIN 18

//extruder pins
#define EXTRUDER_ENCODER_A_PIN 2		//quadrature a pin
#define EXTRUDER_ENCODER_B_PIN 3		//quadrature b pin
#define EXTRUDER_MOTOR_DIR_PIN     4
#define EXTRUDER_MOTOR_SPEED_PIN   5
#define EXTRUDER_HEATER_PIN        6
#define EXTRUDER_FAN_PIN           11
#define EXTRUDER_THERMISTOR_PIN    5  //a -1 disables thermistor readings
#define EXTRUDER_THERMOCOUPLE_PIN  -1 //a -1 disables thermocouple readings
*/

//
// Start of temperature lookup table
//
// Thermistor lookup table for RepRap Temperature Sensor Boards (http://make.rrrf.org/ts)
// Made with createTemperatureLookup.py (http://svn.reprap.org/trunk/reprap/firmware/Arduino/utilities/createTemperatureLookup.py)
// ./createTemperatureLookup.py --r0=100000 --t0=25 --r1=0 --r2=4700 --beta=4066 --max-adc=1023
// r0: 100000
// t0: 25
// r1: 0
// r2: 4700
// beta: 4066
// max adc: 1023
#define NUMTEMPS 20
short temptable[NUMTEMPS][2] = {
   {1, 841},
   {54, 255},
   {107, 209},
   {160, 184},
   {213, 166},
   {266, 153},
   {319, 142},
   {372, 132},
   {425, 124},
   {478, 116},
   {531, 108},
   {584, 101},
   {637, 93},
   {690, 86},
   {743, 78},
   {796, 70},
   {849, 61},
   {902, 50},
   {955, 34},
   {1008, 3}
};
//
// End of temperature lookup table
//

#define EXTRUDER_FORWARD 1
#define EXTRUDER_REVERSE 0

//these our the default values for the extruder.
int extruder_target_celsius = 0;
int extruder_max_celsius = 0;
byte extruder_heater_low = 64;
byte extruder_heater_high = 255;

//this is for doing encoder based extruder control
int extruder_rpm = 0;
long extruder_delay = 0;

//these are for the extruder PID
volatile int extruder_error = 0;						// extruder position / error variable.
volatile int extruder_pGain = EXTRUDER_INITIAL_PGAIN;	// Proportional gain
volatile int extruder_iGain = EXTRUDER_INITIAL_IGAIN;	// Integral gain
volatile int extruder_dGain = EXTRUDER_INITIAL_DGAIN;	// Derivative gain
volatile int iMax = 500;								// Integrator max
volatile int iMin = -500;								// Integrator min
volatile int iState = 0;								// Integrator state
volatile int dState = 0;								// Last position input

// a quick and dirty check to see if our thermistor fell off
int extruder_last_temp=0;
unsigned long extruder_last_temp_time=0;

#ifdef INVERT_EXTRUDER_QUADRATURE
#define EXTRUDER_QUADRATURE_INCREMENT extruder_error--;
#define EXTRUDER_QUADRATURE_DECREMENT extruder_error++;
#else
#define EXTRUDER_QUADRATURE_INCREMENT extruder_error++;
#define EXTRUDER_QUADRATURE_DECREMENT extruder_error--;
#endif

void extruder_read_quadrature()
{  
  // found a low-to-high on channel A
  if (digitalRead(EXTRUDER_ENCODER_A_PIN) == HIGH)
  {   
    // check channel B to see which way
    if (digitalRead(EXTRUDER_ENCODER_B_PIN) == LOW)
        EXTRUDER_QUADRATURE_INCREMENT
    else
        EXTRUDER_QUADRATURE_DECREMENT
  }
  // found a high-to-low on channel A
  else                                        
  {
    // check channel B to see which way
    if (digitalRead(EXTRUDER_ENCODER_B_PIN) == LOW)
        EXTRUDER_QUADRATURE_DECREMENT
    else
        EXTRUDER_QUADRATURE_INCREMENT
  }
}

void init_extruder()
{
	//default to room temp.
	extruder_set_temperature(21);
	
	//setup our motor dir pin
	pinMode(EXTRUDER_MOTOR_DIR_PIN, OUTPUT);
	digitalWrite(EXTRUDER_MOTOR_DIR_PIN, EXTRUDER_FORWARD);

	//setup our motor speed pin
	pinMode(EXTRUDER_MOTOR_SPEED_PIN, OUTPUT);
	analogWrite(EXTRUDER_MOTOR_SPEED_PIN, 0);

	//set up our header pin
	pinMode(EXTRUDER_HEATER_PIN, OUTPUT);
	analogWrite(EXTRUDER_HEATER_PIN, 0);

	//setup our fan pin
	pinMode(EXTRUDER_FAN_PIN, OUTPUT);
	analogWrite(EXTRUDER_FAN_PIN, 0);
	
	//setup our encoder interrupt stuff.
	//these pins are inputs
	pinMode(EXTRUDER_ENCODER_A_PIN, INPUT);
	pinMode(EXTRUDER_ENCODER_B_PIN, INPUT);

	//turn on internal pullups
	digitalWrite(EXTRUDER_ENCODER_A_PIN, HIGH);
	digitalWrite(EXTRUDER_ENCODER_A_PIN, HIGH);
	
	//attach our interrupt handler
	attachInterrupt(0, extruder_read_quadrature, CHANGE);

	//setup our timer interrupt stuff
	setupTimer1Interrupt();
	setupTimer2Interrupt();
	
	//our default speed for the PID is every millisecond (16000 ticks)
	setTimer2Ticks(16000);
}

void extruder_set_temperature(int temp)
{
	extruder_target_celsius = temp;
	extruder_max_celsius = (int)((float)temp * 1.1);
}

/**
*  Samples the temperature and converts it to degrees celsius.
*  Returns degrees celsius.
*/
int extruder_get_temperature()
{
	if (EXTRUDER_THERMISTOR_PIN > -1)
		return extruder_read_thermistor();
	else if (EXTRUDER_THERMOCOUPLE_PIN > -1)
		return extruder_read_thermocouple();
}

/*
* This function gives us the temperature from the thermistor in Celsius
*/
int extruder_read_thermistor()
{
	int raw = extruder_sample_temperature(EXTRUDER_THERMISTOR_PIN);

	int celsius = 0;
	byte i;

	for (i=1; i<NUMTEMPS; i++)
	{
		if (temptable[i][0] > raw)
		{
			celsius  = temptable[i-1][1] + 
				(raw - temptable[i-1][0]) * 
				(temptable[i][1] - temptable[i-1][1]) /
				(temptable[i][0] - temptable[i-1][0]);

			if (celsius > 255)
				celsius = 255; 

			break;
		}
	}

	// Overflow: We just clamp to 0 degrees celsius
	if (i == NUMTEMPS)
		celsius = 0;

	return celsius;
}

/*
* This function gives us the temperature from the thermocouple in Celsius
*/
int extruder_read_thermocouple()
{
	return ( 5.0 * extruder_sample_temperature(EXTRUDER_THERMOCOUPLE_PIN) * 100.0) / 1024.0;
}

/*
* This function gives us an averaged sample of the analog temperature pin.
*/
int extruder_sample_temperature(byte pin)
{
	int raw = 0;
	
	//read in a certain number of samples
	for (byte i=0; i<TEMPERATURE_SAMPLES; i++)
		raw += analogRead(pin);
		
	//average the samples
	raw = raw/TEMPERATURE_SAMPLES;

	//send it back.
	return raw;
}


/*!
  Manages motor and heater based on measured temperature:
  o If temp is too low, don't start the motor
  o Adjust the heater power to keep the temperature at the target
 */
void extruder_manage_temperature()
{
	//make sure we know what our temp is.
	int current_celsius = extruder_get_temperature();

	//put the heater into high mode if we're not at our target.
	if (current_celsius < extruder_target_celsius)
                {
		analogWrite(EXTRUDER_HEATER_PIN, extruder_heater_high);
                //check to see if the temp is actually rising, if not, estop.
                if ((millis() - extruder_last_temp_time) > 10000)  //has it been more than 10 seconds since we last checked the temp?
                     {
                    if ((current_celsius - extruder_last_temp) < 1)  //has there been no temp change AND the heater has been on the whole time?  Shut down!
                        {
                            Serial.print("Temp Sensor Failure. T=");
                            Serial.println(current_celsius,DEC);
                            Serial.println("Reset Arduino to continue.");
                            
                            
                            // Turn heater off
                            extruder_set_temperature(0);
                            analogWrite(EXTRUDER_HEATER_PIN, 0);
                            
                            //  Turn Extruder motor off
                            disableTimer1Interrupt();
			    disableTimer2Interrupt();
			    analogWrite(EXTRUDER_MOTOR_SPEED_PIN, 0);
			    extruder_error = 0;
                            
                            //now what?  wait till someone hits the reset.  Improvements appreciated.
                            while (1)
                              delay(10000);
                        }
                    else
                        extruder_last_temp = current_celsius;
                        extruder_last_temp_time = millis();
                    }
                }

	//put the heater on low if we're at our target.
	else if (current_celsius < extruder_max_celsius)
                {
		analogWrite(EXTRUDER_HEATER_PIN, extruder_heater_low);
                extruder_last_temp=0;  //reset our broken temp sensor check
                }
	//turn the heater off if we're above our max.
	else
		
                {
                analogWrite(EXTRUDER_HEATER_PIN, 0);
                extruder_last_temp=0;    //reset our broken temp sensor check
                }
}

// our point structure to make things nice.
struct LongPoint {
	long x;
	long y;
 	long z;
};

struct FloatPoint {
	float x;
	float y;
 	float z;
};

FloatPoint current_units;
FloatPoint target_units;
FloatPoint delta_units;

FloatPoint current_steps;
FloatPoint target_steps;
FloatPoint delta_steps;

boolean abs_mode = false;   //0 = incremental; 1 = absolute

//default to mm for units
float x_units = X_STEPS_PER_MM;
float y_units = Y_STEPS_PER_MM;
float z_units = Z_STEPS_PER_MM;
float curve_section = CURVE_SECTION_MM;

//our direction vars
byte x_direction = 1;
byte y_direction = 1;
byte z_direction = 1;



//init our string processing
void init_process_string()
{
	//init our command
	for (byte i=0; i<COMMAND_SIZE; i++)
		word[i] = 0;
	serial_count = 0;
}

//our feedrate variables.
float feedrate = 0.0;
long feedrate_micros = 0;

//dwell delay variable.
unsigned long dwelltime = 0;

//Read the string and execute instructions
void process_string(char instruction[], int size)
{
	//the character / means delete block... used for comments and stuff.
	if (instruction[0] == '/')
	{
		Serial.println("ok");
		return;
	}

	//init baby!
	FloatPoint fp;
	fp.x = 0.0;
	fp.y = 0.0;
	fp.z = 0.0;

	byte code = 0;;
	
//what line are we at?
//	long line = -1;
//	if (has_command('N', instruction, size))
//		line = (long)search_string('N', instruction, size);
	
/*
	Serial.print("line: ");
	Serial.println(line);
	Serial.println(instruction);
*/		
	//did we get a gcode?
	if (
		has_command('G', instruction, size) ||
		has_command('X', instruction, size) ||
		has_command('Y', instruction, size) ||
		has_command('Z', instruction, size)
	)
	{
		//which one?
		code = (int)search_string('G', instruction, size);
		
		// Get co-ordinates if required by the code type given
		switch (code)
		{
			case 0:
			case 1:
			case 2:
			case 3:
				if(abs_mode)
				{
					//we do it like this to save time. makes curves better.
					//eg. if only x and y are specified, we dont have to waste time looking up z.
					if (has_command('X', instruction, size))
						fp.x = search_string('X', instruction, size);
					else
						fp.x = current_units.x;
				
					if (has_command('Y', instruction, size))
						fp.y = search_string('Y', instruction, size);
					else
						fp.y = current_units.y;
				
					if (has_command('Z', instruction, size))
						fp.z = search_string('Z', instruction, size);
					else
						fp.z = current_units.z;
				}
				else
				{
					fp.x = search_string('X', instruction, size) + current_units.x;
					fp.y = search_string('Y', instruction, size) + current_units.y;
					fp.z = search_string('Z', instruction, size) + current_units.z;
				}
			break;
		}

		//do something!
		switch (code)
		{
			//Rapid Positioning
			//Linear Interpolation
			//these are basically the same thing.
			case 0:
			case 1:

				//set our target.
				set_target(fp.x, fp.y, fp.z);
			
				//split up xy and z moves
				if (delta_steps.z && (delta_steps.x || delta_steps.y))
				{
					set_target(current_units.x, current_units.y, fp.z);
					figure_feedrate(instruction, size, code);
					dda_move(feedrate_micros);
				}
				
				//do the last move
				set_target(fp.x, fp.y, fp.z);
				figure_feedrate(instruction, size, code);
				dda_move(feedrate_micros);

			break;

			//Dwell
			case 4:
                                //get the time
                                
                                dwelltime = millis() + ((unsigned long)search_string('P', instruction, size));
                             
                                //check the time, and if it's going to be longer than 100 ms, manage the extruder temp
                                
                                while ((dwelltime - millis()) > 100)
                                {
                                      extruder_manage_temperature();
                                }
                                
                                delay (dwelltime-millis());
      
			break;

			//Inches for Units
			case 20:
				x_units = X_STEPS_PER_INCH;
				y_units = Y_STEPS_PER_INCH;
				z_units = Z_STEPS_PER_INCH;
				curve_section = CURVE_SECTION_INCHES;
				
				calculate_deltas();
			break;

			//mm for Units
			case 21:
				x_units = X_STEPS_PER_MM;
				y_units = Y_STEPS_PER_MM;
				z_units = Z_STEPS_PER_MM;
				curve_section = CURVE_SECTION_MM;
				
				calculate_deltas();
			break;

			//go home.
			case 28:
				set_target(0.0, 0.0, 0.0);
				dda_move(getMaxSpeed());
			break;

			//go home via an intermediate point.
			case 30:
				fp.x = search_string('X', instruction, size);
				fp.y = search_string('Y', instruction, size);
				fp.z = search_string('Z', instruction, size);

				//set our target.
				if(abs_mode)
				{
					if (!has_command('X', instruction, size))
						fp.x = current_units.x;
					if (!has_command('Y', instruction, size))
						fp.y = current_units.y;
					if (!has_command('Z', instruction, size))
						fp.z = current_units.z;
						
					set_target(fp.x, fp.y, fp.z);
				}
				else
					set_target(current_units.x + fp.x, current_units.y + fp.y, current_units.z + fp.z);
				
				//go there.
				dda_move(getMaxSpeed());

				//go home.
				set_target(0.0, 0.0, 0.0);
				dda_move(getMaxSpeed());
			break;

			//Absolute Positioning
			case 90:
				abs_mode = true;
			break;

			//Incremental Positioning
			case 91:
				abs_mode = false;
			break;

			//Set as home
			case 92:
				if (has_command('X', instruction, size))
					fp.x = search_string('X', instruction, size);
				else
					fp.x = current_units.x;  
                                        //fp.x=0;
			
				if (has_command('Y', instruction, size))
					fp.y = search_string('Y', instruction, size);
				else
					fp.y = current_units.y;   
                                        //fp.y=0;
			
				if (has_command('Z', instruction, size))
					fp.z = search_string('Z', instruction, size);
				else
					fp.z = current_units.z;   
                                        //fp.z=0;

				set_position(fp.x, fp.y, fp.z);
			break;

/*
			//Inverse Time Feed Mode
			case 93:

			break;  //TODO: add this

			//Feed per Minute Mode
			case 94:

			break;  //TODO: add this
*/

			default:
				Serial.print("huh? G");
				Serial.println(code,DEC);
		}
	}
	
	//find us an m code.
	if (has_command('M', instruction, size))
	{
		code = search_string('M', instruction, size);
		switch (code)
		{
			//TODO: this is a bug because search_string returns 0.  gotta fix that.
			case 0:
				true;
			break;
	/*
			case 0:
				//todo: stop program
			break;

			case 1:
				//todo: optional stop
			break;

			case 2:
				//todo: program end
			break;
	*/		
			//turn extruder on, forward
			case 101:
				extruder_error = 0;
				enableTimer1Interrupt();
				enableTimer2Interrupt();
			break;

			//turn extruder on, reverse
			case 102:
				//this is not implemented, nor supported
			break;

			//turn extruder off
			case 103:
				disableTimer1Interrupt();
				disableTimer2Interrupt();
				analogWrite(EXTRUDER_MOTOR_SPEED_PIN, 0);
				extruder_error = 0;
			break;

			//custom code for temperature control
			case 104:
				extruder_set_temperature((int)search_string('S', instruction, size));

				//warmup if we're too cold.
				/*while (extruder_get_temperature() < extruder_target_celsius)
				{
					extruder_manage_temperature();
					//Serial.print("T:");
					//Serial.println(extruder_get_temperature());
					delay(1000);	
				}*/
				
			break;

			//custom code for temperature reading
			case 105:
				Serial.print("T:");
				Serial.println(extruder_get_temperature());
			break;
			
			//turn fan on
			case 106:
				digitalWrite(EXTRUDER_FAN_PIN, HIGH);
			break;

			//turn fan off
			case 107:
				digitalWrite(EXTRUDER_FAN_PIN, LOW);
			break;
			
			//set max extruder speed, in RPM
			case 108:
				extruder_rpm = (int)search_string('S', instruction, size);

				if (extruder_rpm)
				{
					extruder_delay = (960000000UL / EXTRUDER_ENCODER_STEPS) / extruder_rpm;
					setTimer1Ticks(extruder_delay);
				}
			break;
			
			case 110:
				Serial.println("Hi Skeinforge.");
				break;			

			//set extruder P gain
			case 120:
				extruder_pGain = (int)search_string('S', instruction, size);
			break;

			//set extruder I gain
			case 121:
				extruder_iGain = (int)search_string('S', instruction, size);
			break;

			//set extruder D gain
			case 122:
				extruder_dGain = (int)search_string('S', instruction, size);
			break;

			//set extruder iMax
			case 123:
				iMax = (int)search_string('S', instruction, size);
			break;

			//set extruder iMin
			case 124:
				iMin = (int)search_string('S', instruction, size);
			break;
			
			//set extruder iMin
			case 125:
				Serial.print("P:");
				Serial.println(extruder_pGain, DEC);
				Serial.print("I:");
				Serial.println(extruder_iGain, DEC);
				Serial.print("D:");
				Serial.println(extruder_dGain, DEC);
			break;
			
			//open the valve
			case 126:
				digitalWrite(EXTRUDER_VALVE_PIN, LOW);
			break;

			//open the valve
			case 127:
				digitalWrite(EXTRUDER_VALVE_PIN, HIGH);
			break;
			
			default:
				Serial.print("Huh? M");
				Serial.println(code);
		}		
	}
	
	//tell our host we're done.
	Serial.println("ok");
}

//look for the number that appears after the char key and return it
double search_string(char key, char instruction[], int string_size)
{
	char temp[10] = "";

	for (byte i=0; i<string_size; i++)
	{
		if (instruction[i] == key)
		{
			i++;      
			int k = 0;
			while (i < string_size && k < 10)
			{
				if (instruction[i] == 0 || instruction[i] == ' ')
					break;

				temp[k] = instruction[i];
				i++;
				k++;
			}
			return strtod(temp, NULL);
		}
	}
	
	return 0;
}

//look for the command if it exists.
bool has_command(char key, char instruction[], int string_size)
{
	for (byte i=0; i<string_size; i++)
	{
		if (instruction[i] == key)
			return true;
	}
	
	return false;
}

void figure_feedrate(char instruction[], int size, byte code)
{
	//do we have a set speed?
	if (has_command('G', instruction, size))
	{
		//adjust if we have a specific feedrate.
		if (code == 1)
		{
			//how fast do we move?
			feedrate = search_string('F', instruction, size);
			if (feedrate > 0)
				feedrate_micros = calculate_feedrate_delay(feedrate);
			//nope, no feedrate
			else
				feedrate_micros = getMaxSpeed();
		}
		//use our max for normal moves.
		else
			feedrate_micros = getMaxSpeed();
	}
	//nope, just coordinates!
	else
	{
		//do we have a feedrate yet?
		if (feedrate > 0)
			feedrate_micros = calculate_feedrate_delay(feedrate);
		//nope, no feedrate
		else
			feedrate_micros = getMaxSpeed();
	}
}


//init our variables
long max_delta;
long x_counter;
long y_counter;
long z_counter;
bool x_can_step;
bool y_can_step;
bool z_can_step;
int milli_delay;

void init_steppers()
{
	//turn them off to start.
	disable_steppers();
	
	//init our points.
	current_units.x = 0.0;
	current_units.y = 0.0;
	current_units.z = 0.0;
	target_units.x = 0.0;
	target_units.y = 0.0;
	target_units.z = 0.0;
	
	pinMode(X_STEP_PIN, OUTPUT);
	pinMode(X_DIR_PIN, OUTPUT);
	pinMode(X_ENABLE_PIN, OUTPUT);
	pinMode(X_MIN_PIN, INPUT);
	pinMode(X_MAX_PIN, INPUT);
	
	pinMode(Y_STEP_PIN, OUTPUT);
	pinMode(Y_DIR_PIN, OUTPUT);
	pinMode(Y_ENABLE_PIN, OUTPUT);
	pinMode(Y_MIN_PIN, INPUT);
	pinMode(Y_MAX_PIN, INPUT);
	
	pinMode(Z_STEP_PIN, OUTPUT);
	pinMode(Z_DIR_PIN, OUTPUT);
	pinMode(Z_ENABLE_PIN, OUTPUT);
	pinMode(Z_MIN_PIN, INPUT);
	pinMode(Z_MAX_PIN, INPUT);
	
	//figure our stuff.
	calculate_deltas();
}

void dda_move(long micro_delay)
{
	//enable our steppers
	digitalWrite(X_ENABLE_PIN, HIGH);
	digitalWrite(Y_ENABLE_PIN, HIGH);
	digitalWrite(Z_ENABLE_PIN, HIGH);
	
	//figure out our deltas
	max_delta = max(delta_steps.x, delta_steps.y);
	max_delta = max(delta_steps.z, max_delta);

	//init stuff.
	long x_counter = -max_delta/2;
	long y_counter = -max_delta/2;
	long z_counter = -max_delta/2;
	
	//our step flags
	bool x_can_step = 0;
	bool y_can_step = 0;
	bool z_can_step = 0;
	
	if (micro_delay >= 16383)
		milli_delay = micro_delay / 1000;
	else
		milli_delay = 0;

/*	
	Serial.print("max:");
	Serial.println(max_delta, DEC);
	Serial.print("xd:");
	Serial.println(delta_steps.x, DEC);
	Serial.print("yd:");
	Serial.println(delta_steps.y, DEC);
	Serial.print("zd:");
	Serial.println(delta_steps.z, DEC);
	
	Serial.print("msec:");
	Serial.println(millis, DEC);
	Serial.print("usec:");
	Serial.println(micro_delay, DEC);
*/
	//do our DDA line!
	do
	{
		x_can_step = can_step(X_MIN_PIN, X_MAX_PIN, current_steps.x, target_steps.x, x_direction);
		y_can_step = can_step(Y_MIN_PIN, Y_MAX_PIN, current_steps.y, target_steps.y, y_direction);
		z_can_step = can_step(Z_MIN_PIN, Z_MAX_PIN, current_steps.z, target_steps.z, z_direction);

		if (x_can_step)
		{
			x_counter += delta_steps.x;
			
			if (x_counter > 0)
			{
				do_step(X_STEP_PIN);
				x_counter -= max_delta;
				
				if (x_direction)
					current_steps.x++;
				else
					current_steps.x--;
			}
		}

		if (y_can_step)
		{
			y_counter += delta_steps.y;
			
			if (y_counter > 0)
			{
				do_step(Y_STEP_PIN);
				y_counter -= max_delta;

				if (y_direction)
					current_steps.y++;
				else
					current_steps.y--;
			}
		}
		
		if (z_can_step)
		{
			z_counter += delta_steps.z;
			
			if (z_counter > 0)
			{
				do_step(Z_STEP_PIN);
				z_counter -= max_delta;
				
				if (z_direction)
					current_steps.z++;
				else
					current_steps.z--;
			}
		}
		
		extruder_manage_temperature();
				
		//wait for next step.
		if (milli_delay > 0)
			delay(milli_delay);			
		else
			delayMicrosecondsInterruptible(micro_delay);
	}
	while (x_can_step || y_can_step || z_can_step);
	
	//set our points to be the same
	current_units.x = target_units.x;
	current_units.y = target_units.y;
	current_units.z = target_units.z;
	calculate_deltas();
}

bool can_step(byte min_pin, byte max_pin, long current, long target, byte direction)
{
	//stop us if we're on target
	if (target == current)
		return false;
	//stop us if we're at home and still going 
	else if (read_switch(min_pin) && !direction)
		return false;
	//stop us if we're at max and still going
	else if (read_switch(max_pin) && direction)
		return false;

	//default to being able to step
	return true;
}

void do_step(byte step_pin)
{
	digitalWrite(step_pin, HIGH);
	delayMicrosecondsInterruptible(5);
	digitalWrite(step_pin, LOW);
}

bool read_switch(byte pin)
{
	//dual read as crude debounce
	
	if ( SENSORS_INVERTING )
		return !digitalRead(pin) && !digitalRead(pin);
	else
		return digitalRead(pin) && digitalRead(pin);
}

long to_steps(float steps_per_unit, float units)
{
	return steps_per_unit * units;
}

void set_target(float x, float y, float z)
{
	target_units.x = x;
	target_units.y = y;
	target_units.z = z;
	
	calculate_deltas();
}

void set_position(float x, float y, float z)
{
	current_units.x = x;
	current_units.y = y;
	current_units.z = z;
	
	calculate_deltas();
}

void calculate_deltas()
{
	//figure our deltas.
	delta_units.x = abs(target_units.x - current_units.x);
	delta_units.y = abs(target_units.y - current_units.y);
	delta_units.z = abs(target_units.z - current_units.z);
				
	//set our steps current, target, and delta
	current_steps.x = to_steps(x_units, current_units.x);
	current_steps.y = to_steps(y_units, current_units.y);
	current_steps.z = to_steps(z_units, current_units.z);

	target_steps.x = to_steps(x_units, target_units.x);
	target_steps.y = to_steps(y_units, target_units.y);
	target_steps.z = to_steps(z_units, target_units.z);

	delta_steps.x = abs(target_steps.x - current_steps.x);
	delta_steps.y = abs(target_steps.y - current_steps.y);
	delta_steps.z = abs(target_steps.z - current_steps.z);
	
	//what is our direction
	x_direction = (target_units.x >= current_units.x);
	y_direction = (target_units.y >= current_units.y);
	z_direction = (target_units.z >= current_units.z);

	//set our direction pins as well
	digitalWrite(X_DIR_PIN, x_direction);
	digitalWrite(Y_DIR_PIN, y_direction);
	digitalWrite(Z_DIR_PIN, z_direction);
}


long calculate_feedrate_delay(float feedrateW)
{
	if ((delta_steps.x || delta_steps.y) && feedrateW > FAST_XY_FEEDRATE)
		feedrateW = FAST_XY_FEEDRATE;
	if ((delta_steps.z>0) && (feedrateW > FAST_Z_FEEDRATE))
                feedrateW = FAST_Z_FEEDRATE;

	//how long is our line length?
	float distance = sqrt(delta_units.x*delta_units.x + delta_units.y*delta_units.y + delta_units.z*delta_units.z);
	long master_steps = 0;
	
	//find the dominant axis.
	if (delta_steps.x > delta_steps.y)
	{
		if (delta_steps.z > delta_steps.x)
			master_steps = delta_steps.z;
		else
			master_steps = delta_steps.x;
	}
	else
	{
		if (delta_steps.z > delta_steps.y)
			master_steps = delta_steps.z;
		else
			master_steps = delta_steps.y;
	}

	//calculate delay between steps in microseconds.  this is sort of tricky, but not too bad.
	//the formula has been condensed to save space.  here it is in english:
	// distance / feedrate * 60000000.0 = move duration in microseconds
	// move duration / master_steps = time between steps for master axis.

	return ((distance * 60000000.0) / feedrateW) / master_steps;	
}

long getMaxSpeed()
{
	if (delta_steps.z > 0)
		return calculate_feedrate_delay(FAST_Z_FEEDRATE);
	else
		return calculate_feedrate_delay(FAST_XY_FEEDRATE);
}

void disable_steppers()
{
	//enable our steppers
	digitalWrite(X_ENABLE_PIN, LOW);
	digitalWrite(Y_ENABLE_PIN, LOW);
	digitalWrite(Z_ENABLE_PIN, LOW);
}

//these routines provide an easy interface for controlling timer1 interrupts

//this handles the timer interrupt event
SIGNAL(SIG_OUTPUT_COMPARE1A)
{
	extruder_error--;
}

void enableTimer1Interrupt()
{
	//enable our interrupt!
	TIMSK1 |= (1<<OCIE1A);
}

void disableTimer1Interrupt()
{
	TIMSK1 &= ~(1<<ICIE1);
	TIMSK1 &= ~(1<<OCIE1A);
}

void setTimer1Resolution(byte r)
{
	//from table 15-5 in that atmega168 datasheet:
	// we're setting CS12 - CS10 which correspond to the binary numbers 0-5
	// 0 = no timer
	// 1 = no prescaler
	// 2 = clock/8
	// 3 = clock/64
	// 4 = clock/256
	// 5 = clock/1024
	
	if (r > 5)
		r = 5;
		
	TCCR1B &= B11111000;
	TCCR1B |= r;
}

void setTimer1Ceiling(unsigned int c)
{
	OCR1A = c;
}


unsigned int getTimer1Ceiling(unsigned long ticks)
{
	// our slowest speed at our highest resolution ( (2^16-1) * 0.0625 usecs = 4095 usecs)
	if (ticks <= 65535L)
		return (ticks & 0xffff);
	// our slowest speed at our next highest resolution ( (2^16-1) * 0.5 usecs = 32767 usecs)
	else if (ticks <= 524280L)
		return ((ticks / 8) & 0xffff);
	// our slowest speed at our medium resolution ( (2^16-1) * 4 usecs = 262140 usecs)
	else if (ticks <= 4194240L)
		return ((ticks / 64) & 0xffff);
	// our slowest speed at our medium-low resolution ( (2^16-1) * 16 usecs = 1048560 usecs)
	else if (ticks <= 16776960L)
		return (ticks / 256);
	// our slowest speed at our lowest resolution ((2^16-1) * 64 usecs = 4194240 usecs)
	else if (ticks <= 67107840L)
		return (ticks / 1024);
	//its really slow... hopefully we can just get by with super slow.
	else
		return 65535;
}

byte getTimer1Resolution(unsigned long ticks)
{
	// these also represent frequency: 1000000 / ticks / 2 = frequency in hz.
	
	// our slowest speed at our highest resolution ( (2^16-1) * 0.0625 usecs = 4095 usecs (4 millisecond max))
	// range: 8Mhz max - 122hz min
	if (ticks <= 65535L)
		return 1;
	// our slowest speed at our next highest resolution ( (2^16-1) * 0.5 usecs = 32767 usecs (32 millisecond max))
	// range:1Mhz max - 15.26hz min
	else if (ticks <= 524280L)
		return 2;
	// our slowest speed at our medium resolution ( (2^16-1) * 4 usecs = 262140 usecs (0.26 seconds max))
	// range: 125Khz max - 1.9hz min
	else if (ticks <= 4194240L)
		return 3;
	// our slowest speed at our medium-low resolution ( (2^16-1) * 16 usecs = 1048560 usecs (1.04 seconds max))
	// range: 31.25Khz max - 0.475hz min
	else if (ticks <= 16776960L)
		return 4;
	// our slowest speed at our lowest resolution ((2^16-1) * 64 usecs = 4194240 usecs (4.19 seconds max))
	// range: 7.812Khz max - 0.119hz min
	else if (ticks <= 67107840L)
		return 5;
	//its really slow... hopefully we can just get by with super slow.
	else
		return 5;
}

void setTimer1Ticks(unsigned long ticks)
{
	// ticks is the delay between interrupts in 4 microsecond ticks.
	//
	// we break it into 5 different resolutions based on the delay. 
	// then we set the resolution based on the size of the delay.
	// we also then calculate the timer ceiling required. (ie what the counter counts to)
	// the result is the timer counts up to the appropriate time and then fires an interrupt.

	setTimer1Ceiling(getTimer1Ceiling(ticks));
	setTimer1Resolution(getTimer1Resolution(ticks));
}

void setupTimer1Interrupt()
{
	//clear the registers
	TCCR1A = 0;
	TCCR1B = 0;
	TCCR1C = 0;
	TIMSK1 = 0;
	
	//waveform generation = 0100 = CTC
	TCCR1B &= ~(1<<WGM13);
	TCCR1B |=  (1<<WGM12);
	TCCR1A &= ~(1<<WGM11); 
	TCCR1A &= ~(1<<WGM10);

	//output mode = 00 (disconnected)
	TCCR1A &= ~(1<<COM1A1); 
	TCCR1A &= ~(1<<COM1A0);
	TCCR1A &= ~(1<<COM1B1); 
	TCCR1A &= ~(1<<COM1B0);

	//start off with a slow frequency.
	setTimer1Resolution(5);
	setTimer1Ceiling(65535);
	disableTimer1Interrupt();
}

//these routines provide an easy interface for controlling Timer2 interrupts
volatile int cnt = 0;
volatile int foo2 = 0;

//this handles the timer interrupt event
SIGNAL(SIG_OUTPUT_COMPARE2A)
{
  	// somewhat hacked implementation of a PID algorithm as described at:
	// http://www.embedded.com/2000/0010/0010feat3.htm - PID Without a PhD, Tim Wescott 
	
	int abs_error = abs(extruder_error);
	int pTerm = 0;
	int iTerm = 0;
	int dTerm = 0;
	int speed = 0;

/*
	//THIS WAS WAY MORE TROUBLE THAN IT WAS WORTH. ABORTED PRINTS 3/4 OF THE WAY THROUGH. ARGH.
	
	//if our error is too high, it means we cant keep up.  bail to protect the extruder motor
	if (abs_error > 750)
	{
		disableTimer1Interrupt();
		disableTimer2Interrupt();
		extruder_error = 0;
		analogWrite(EXTRUDER_MOTOR_SPEED_PIN, 0);
		Serial.println("Extruder Fail");
	}
	else
	{
*/
		//hack for extruder not keeping up, overflowing, then shutting off.
		if (extruder_error < -5000)
			extruder_error = -500;
		if (extruder_error > 5000)
			extruder_error = 500;

		if (extruder_error < 0)
		{
			//calculate our P term
                        // since we're using ints here, I switched the divide to a multiply, and I'm using 10 as my pgain.  wbortz@gmail.com
			// pTerm = abs_error / extruder_pGain;
			pTerm = abs_error * extruder_pGain;	

			//calculate our I term
			iState += abs_error;
			iState = constrain(iState, iMin, iMax);
			iTerm = iState / extruder_iGain;
	
			//calculate our D term
			dTerm = (abs_error - dState) * extruder_dGain;
			dState = abs_error;

			//calculate our PWM, within bounds.
			speed = pTerm + iTerm - dTerm;
		}

		//our debug loop checker thingie
		/*
		cnt++;
		if (cnt > 250)
		{
			//Serial.print("e:");
			Serial.print(extruder_error);
			Serial.print(", ");
			Serial.println(speed);
			cnt = 0;
		}
		*/

		//figure out our real speed and use it.
		speed = constrain(speed, EXTRUDER_MIN_SPEED, EXTRUDER_MAX_SPEED);

                //debug
                //if (foo2 > 1000)
                //{
                //  foo2 = 0;
                //  Serial.print(extruder_error);
                //  Serial.print("/");
                //  Serial.println(speed);
                //}
                foo2++;

		analogWrite(EXTRUDER_MOTOR_SPEED_PIN, speed);
//	}
}
/*
SIGNAL(SIG_OUTPUT_COMPARE2A)
{
	// straight-forward implementation of a PID algorithm as described at:
	// http://www.embedded.com/2000/0010/0010feat3.htm - PID Without a PhD, Tim Wescott 

	float pTerm;
	float iTerm;
	float dTerm;
	int speed;
	int abs_error = abs(extruder_error);

	//if our error is too high, it means we cant keep up.  bail to protect the extruder motor
	if (abs_error > 1000)
	{
		disableTimer1Interrupt();
		disableTimer2Interrupt();
		extruder_error = 0;
		analogWrite(EXTRUDER_MOTOR_SPEED_PIN, 0);
		Serial.println("Extruder Fail");
	}
	else
	{
		//calculate our P term
		pTerm = extruder_pGain * (float)abs_error;
	
		//calculate our I term
		iState += -extruder_error;
		if (iState > iMax)
			iState = iMax;
		else if (iState < iMin)
			iState = iMin;
		iTerm = extruder_iGain * (float)iState;
	
		//calculate our D term
		dTerm = extruder_dGain * (float)(abs_error - dState);
		dState = abs_error;

		//calculate our PWM
		int speed = ceil(pTerm + iTerm - dTerm);

		//do some bounds checking.
		speed = max(speed, EXTRUDER_MIN_SPEED);
		speed = min(speed, EXTRUDER_MAX_SPEED);

		//our debug loop.
		cnt++;
		if (cnt > 250)
		{
			Serial.print("e:");
			Serial.println(extruder_error);
			Serial.print("spd:");
			Serial.println(speed);
			cnt = 0;
		}

		digitalWrite(EXTRUDER_MOTOR_DIR_PIN, EXTRUDER_FORWARD);
		analogWrite(EXTRUDER_MOTOR_SPEED_PIN, speed);
	}
}
*/

void enableTimer2Interrupt()
{
	//enable our interrupt!
	TIMSK2 |= (1<<OCIE2A);
}

void disableTimer2Interrupt()
{
	TIMSK2 &= ~(1<<OCIE2A);
}

void setTimer2Resolution(byte r)
{
	//from table 17-9 in that atmega168 datasheet:
	// we're setting CS22 - CS20 which correspond to the binary numbers 0-7
	// 0 = no timer
	// 1 = no prescaler
	// 2 = clock/8
	// 3 = clock/32
	// 4 = clock/64
	// 5 = clock/128
	// 6 = clock/256
	// 7 = clock/1024
	
	if (r > 7)
		r = 7;
		
	TCCR2B &= B11111000;
	TCCR2B |= r;
}

void setTimer2Ceiling(byte c)
{
	OCR2A = c;
}


byte getTimer2Ceiling(unsigned long ticks)
{
	if (ticks <= 255L)
		return (ticks & 0xff);
	else if (ticks <= 2040L)
		return ((ticks / 8) & 0xff);
	else if (ticks <= 8160L)
		return ((ticks / 32) & 0xff);
	else if (ticks <= 16320L)
		return ((ticks / 64) & 0xff);
	else if (ticks <= 32640L)
		return ((ticks / 128) & 0xff);
	else if (ticks <= 65280L)
		return ((ticks / 256) & 0xff);
	else if (ticks <= 261120L)
		return ((ticks / 1024) & 0xff);
	else
		return 255;
}

byte getTimer2Resolution(unsigned long ticks)
{
	if (ticks <= 255L)
		return 1;
	else if (ticks <= 2040L)
		return 2;
	else if (ticks <= 8160L)
		return 3;
	else if (ticks <= 16320L)
		return 4;
	else if (ticks <= 32640L)
		return 5;
	else if (ticks <= 65280L)
		return 6;
	else if (ticks <= 261120L)
		return 7;
	else
		return 7;
}

void setTimer2Ticks(unsigned long ticks)
{
	// ticks is the delay between interrupts in 4 microsecond ticks.
	//
	// we break it into 5 different resolutions based on the delay. 
	// then we set the resolution based on the size of the delay.
	// we also then calculate the timer ceiling required. (ie what the counter counts to)
	// the result is the timer counts up to the appropriate time and then fires an interrupt.

	setTimer2Ceiling(getTimer2Ceiling(ticks));
	setTimer2Resolution(getTimer2Resolution(ticks));
}

void setupTimer2Interrupt()
{
	//clear the registers
	TCCR2A = 0;
	TCCR2B = 0;
	TIMSK2 = 0;
	
	//waveform generation = 010 = CTC
	TCCR2B &= ~(1<<WGM22);
	TCCR2A |=  (1<<WGM21); 
	TCCR2A &= ~(1<<WGM20);

	//output mode = 00 (disconnected)
	TCCR2A &= ~(1<<COM2A1); 
	TCCR2A &= ~(1<<COM2A0);
	TCCR2A &= ~(1<<COM2B1); 
	TCCR2A &= ~(1<<COM2B0);

	//start off with a slow frequency.
	setTimer2Resolution(7);
	setTimer2Ceiling(255);
	disableTimer2Interrupt();
}


int main(void)
{
	init();

	setup();
    
	for (;;)
		loop();
        
	return 0;
}

